@@ -1,4 +1,4 @@
-/* dnsmasq is Copyright (c) 2000 - 2004 Simon Kelley
+/* dnsmasq is Copyright (c) 2000 - 2005 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@ struct myoption {
   int val;
 };
 
-#define OPTSTRING "ZDNLERzowefnbvhdkqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:P:"
+#define OPTSTRING "yZDNLERKzowefnbvhdkqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:P:J:W:"
 
 static struct myoption opts[] = { 
   {"version", 0, 0, 'v'},
@@ -72,8 +72,12 @@ static struct myoption opts[] = {
   {"alias", 1, 0, 'V' },
   {"dhcp-vendorclass", 1, 0, 'U'},
   {"dhcp-userclass", 1, 0, 'j'},
+  {"dhcp-ignore", 1, 0, 'J'},
   {"edns-packet-max", 1, 0, 'P'},
   {"keep-in-foreground", 0, 0, 'k'},
+  {"dhcp-authoritative", 0, 0, 'K'},
+  {"srv-host", 1, 0, 'W'},
+  {"localise-queries", 0, 0, 'y'},
   {0, 0, 0, 0}
 };
 
@@ -91,6 +95,7 @@ static struct optflags optmap[] = {
   { 'n', OPT_NO_POLL },
   { 'd', OPT_DEBUG },
   { 'k', OPT_NO_FORK },
+  { 'K', OPT_AUTHORITATIVE },
   { 'o', OPT_ORDER },
   { 'R', OPT_NO_RESOLV },
   { 'E', OPT_EXPAND },
@@ -99,14 +104,18 @@ static struct optflags optmap[] = {
   { 'D', OPT_NODOTS_LOCAL },
   { 'z', OPT_NOWILD },
   { 'Z', OPT_ETHERS },
+  { 'y', OPT_LOCALISE },
   { 'v', 0},
   { 'w', 0},
   { 0, 0 }
 };
 
 static char *usage =
-"Usage: dnsmasq [options]\n"
-"\nValid options are :\n"
+"Usage: dnsmasq [options]\n\n"
+#ifndef HAVE_GETOPT_LONG
+"Use short options only on the command line.\n"
+#endif
+"Valid options are :\n"
 "-a, --listen-address=ipaddr         Specify local address(es) to listen on.\n"
 "-A, --address=/domain/ipaddr        Return ipaddr for all hosts in specified domains.\n"
 "-b, --bogus-priv                    Fake reverse lookups for RFC1918 private address ranges.\n"
@@ -126,10 +135,12 @@ static char *usage =
 "-i, --interface=interface           Specify interface(s) to listen on.\n"
 "-I, --except-interface=int          Specify interface(s) NOT to listen on.\n"
 "-j, --dhcp-userclass=<id>,<class>   Map DHCP user class to option set.\n"
+"-J, --dhcp-ignore=<id>              Don't do DHCP for hosts in option set.\n"
 "-k, --keep-in-foreground            Do NOT fork into the background, do NOT run in debug mode.\n"
+"-K, --dhcp-authoritative            Assume we are the only DHCP server on the local network.\n"
 "-l, --dhcp-leasefile=path           Specify where to store DHCP leases (defaults to " LEASEFILE ").\n"
 "-L, --localmx                       Return MX records for local hosts.\n"
-"-m, --mx-host=host_name             Specify the MX name to reply to.\n"
+"-m, --mx-host=host_name,target,pref Specify an MX record.\n"
 "-M, --dhcp-boot=<bootp opts>        Specify BOOTP options to DHCP server.\n"
 "-n, --no-poll                       Do NOT poll " RESOLVFILE " file, reload only on SIGHUP.\n"
 "-N, --no-negcache                   Do NOT cache failed search results.\n"
@@ -144,15 +155,17 @@ static char *usage =
 "-S, --server=/domain/ipaddr         Specify address(es) of upstream servers with optional domains.\n"
 "    --local=/domain/                Never forward queries to specified domains.\n"
 "-s, --domain=domain                 Specify the domain to be assigned in DHCP leases.\n"
-"-t, --mx-target=host_name           Specify the host in an MX reply.\n"
+"-t, --mx-target=host_name           Specify default target in an MX record.\n"
 "-T, --local-ttl=time                Specify time-to-live in seconds for replies from /etc/hosts.\n"
 "-u, --user=username                 Change to this user after startup. (defaults to " CHUSER ").\n" 
 "-U, --dhcp-vendorclass=<id>,<class> Map DHCP vendor class to option set.\n"
 "-v, --version                       Display dnsmasq version and copyright information.\n"
 "-V, --alias=addr,addr,mask          Translate IPv4 addresses from upstream servers.\n"
+"-W, --srv-host=name,port,pri,weight Specify a SRV record.\n"
 "-w, --help                          Display this message.\n"
 "-x, --pid-file=path                 Specify path of PID file. (defaults to " RUNFILE ").\n"
 "-X, --dhcp-lease-max=number         Specify maximum number of DHCP leases (defaults to %d).\n"
+"-y, --localise-queries              Answer DNS queries based on the interface a query was sent to."
 "-z, --bind-interfaces               Bind only to interfaces in use.\n"
 "-Z, --read-ethers                   Read DHCP static host information from " ETHERSFILE ".\n"
 "\n";
@@ -164,8 +177,8 @@ struct daemon *read_opts (int argc, char **argv)
   char *problem = NULL, *buff = safe_malloc(MAXDNAME);
   int option = 0, i;
   FILE *file_save = NULL, *f = NULL;
-  char *file_name_save = NULL, *conffile = CONFFILE;
-  int conffile_set = 0;
+  char *comma, *file_name_save = NULL, *conffile = CONFFILE;
+  int hosts_index = 1, conffile_set = 0;
   int line_save = 0, lineno = 0;
   opterr = 0;
   
@@ -364,25 +377,41 @@ struct daemon *read_opts (int argc, char **argv)
 
 	    case 'm':
 	      {
-		char *comma = strchr(optarg, ',');
-		if (comma)
-		  *(comma++) = 0;
+		int pref = 1;
+		struct mx_record *new;
+
+		if ((comma = strchr(optarg, ',')))
+		  {
+		    char *prefstr;
+		    *(comma++) = 0;
+		    if ((prefstr=strchr(comma, ',')))
+		      {
+			*(prefstr++) = 0;
+			if (!atoi_check(prefstr, &pref))
+			  {
+			    option = '?';
+			    problem = "bad MX preference";
+			    break;
+			  }
+		      }
+		  }
+
 		if (!canonicalise(optarg) || (comma && !canonicalise(comma)))
 		  {
 		    option = '?';
 		    problem = "bad MX name";
+		    break;
 		  }
-		else 
-		  {
-		    struct mx_record *new = safe_malloc(sizeof(struct mx_record));
-		    new->next = daemon->mxnames;
-		    daemon->mxnames = new;
-		    new->mxname = safe_string_alloc(optarg);
-		    new->mxtarget = safe_string_alloc(comma); /* may be NULL */
-		  }
+
+		new = safe_malloc(sizeof(struct mx_record));
+		new->next = daemon->mxnames;
+		daemon->mxnames = new;
+		new->mxname = safe_string_alloc(optarg);
+		new->mxtarget = safe_string_alloc(comma); /* may be NULL */
+		new->preference = pref;
 		break;
 	      }
-	      
+
 	    case 't':
 	      if (!canonicalise(optarg))
 		{
@@ -398,15 +427,15 @@ struct daemon *read_opts (int argc, char **argv)
 	      break;
 	      
 	    case 'H':
-	      if (daemon->addn_hosts)
-		{
-		  option = '?';
-		  problem = "only one addn hosts file allowed";
-		}
-	      else
-		daemon->addn_hosts = safe_string_alloc(optarg);
-	      break;
-	      
+	      {
+		struct hostsfile *new = safe_malloc(sizeof(struct hostsfile));
+		new->fname = safe_string_alloc(optarg);
+		new->index = hosts_index++;
+		new->next = daemon->addn_hosts;
+		daemon->addn_hosts = new;
+		break;
+	      }
+
 	    case 's':
 	      if (strcmp (optarg, "#") == 0)
 		daemon->options |= OPT_RESOLV_DOMAIN;
@@ -425,8 +454,10 @@ struct daemon *read_opts (int argc, char **argv)
 	      break;
 	      
 	    case 'i':
-	      {
+	      do {
 		struct iname *new = safe_malloc(sizeof(struct iname));
+		if ((comma = strchr(optarg, ',')))
+		  *comma++ = 0;
 		new->next = daemon->if_names;
 		daemon->if_names = new;
 		/* new->name may be NULL if someone does
@@ -435,20 +466,24 @@ struct daemon *read_opts (int argc, char **argv)
 		new->isloop = new->used = 0;
 		if (strchr(optarg, ':'))
 		  daemon->options |= OPT_NOWILD;
-		break;
-	      }
-	      
+		optarg = comma;
+	      } while (optarg);
+	      break;
+	    
 	    case 'I':
-	      {
+	      do {
 		struct iname *new = safe_malloc(sizeof(struct iname));
+		if ((comma = strchr(optarg, ',')))
+		  *comma++ = 0;
 		new->next = daemon->if_except;
 		daemon->if_except = new;
 		new->name = safe_string_alloc(optarg);
 		if (strchr(optarg, ':'))
-		   daemon->options |= OPT_NOWILD;
-		break;
-	      }
-	      
+		  daemon->options |= OPT_NOWILD;
+		optarg = comma;
+	      } while (optarg);
+	      break;
+	      	      
 	    case 'B':
 	      {
 		struct in_addr addr;
@@ -465,8 +500,10 @@ struct daemon *read_opts (int argc, char **argv)
 	      }
 
 	    case 'a':
-	      {
+	      do {
 		struct iname *new = safe_malloc(sizeof(struct iname));
+		if ((comma = strchr(optarg, ',')))
+		  *comma++ = 0;
 		new->next = daemon->if_addrs;
 #ifdef HAVE_IPV6
 		if (inet_pton(AF_INET, optarg, &new->addr.in.sin_addr))
@@ -497,14 +534,14 @@ struct daemon *read_opts (int argc, char **argv)
 		  {
 		    option = '?'; /* error */
 		    free(new);
-		    new = NULL;
+		    break;
 		  }
 		
-		if (new)
-		  daemon->if_addrs = new;
-		break;
-	      }
-	      
+		daemon->if_addrs = new;
+		optarg = comma;
+	      } while (optarg);
+	      break;
+	      	      
 	    case 'S':
 	    case 'A':
 	      {
@@ -722,7 +759,7 @@ struct daemon *read_opts (int argc, char **argv)
 	    case 'F':
 	      {
 		int k, leasepos = 2;
-		char *cp, *comma, *a[5] = { NULL, NULL, NULL, NULL, NULL };
+		char *cp, *a[5] = { NULL, NULL, NULL, NULL, NULL };
 		struct dhcp_context *new = safe_malloc(sizeof(struct dhcp_context));
 		
 		new->next = daemon->dhcp;
@@ -732,7 +769,7 @@ struct daemon *read_opts (int argc, char **argv)
 		new->broadcast.s_addr = 0;
 		new->router.s_addr = 0;
 		new->netid.net = NULL;
-		new->static_only = 0;
+		new->static_only = new->filter_netid = 0;
 		
 		problem = "bad dhcp-range";
 
@@ -743,7 +780,14 @@ struct daemon *read_opts (int argc, char **argv)
 		if (*cp != ',' && (comma = strchr(optarg, ',')))
 		  {
 		    *comma = 0;
-		    new->netid.net = safe_string_alloc(optarg);
+		    if (strstr(optarg, "net:") == optarg)
+		      {
+			new->netid.net = safe_string_alloc(optarg+4);
+			new->netid.next = NULL;
+			new->filter_netid = 1;
+		      }
+		    else
+		      new->netid.net = safe_string_alloc(optarg);
 		    a[0] = comma + 1;
 		  }
 		else
@@ -899,10 +943,7 @@ struct daemon *read_opts (int argc, char **argv)
 			      memcpy(new->clid, arg, len);
 			    }
 			}
-		      else if ((arg[0] == 'n' || arg[0] == 'N') &&
-			       (arg[1] == 'e' || arg[1] == 'E') &&
-			       (arg[2] == 't' || arg[3] == 'T') &&
-			       arg[3] == ':')
+		      else if (strstr(arg, "net:") == arg)
 			{
 			  new->flags |= CONFIG_NETID;
 			  new->netid.net = safe_string_alloc(arg+4);
@@ -1002,165 +1043,274 @@ struct daemon *read_opts (int argc, char **argv)
 	    case 'O':
 	      {
 		struct dhcp_opt *new = safe_malloc(sizeof(struct dhcp_opt));
-		char *cp, *comma;
+		char *cp;
 		int addrs, digs, is_addr, is_hex, is_dec;
 		
 		new->next = daemon->dhcp_opts;
 		new->len = 0;
 		new->is_addr = 0;
 		new->netid = NULL;
+		new->val = NULL;
 				
 		if ((comma = strchr(optarg, ',')))
 		  {
-		    *comma = 0;
+		    struct dhcp_netid *np = NULL;
+		    *comma++ = 0;
 		
-		    for (cp = optarg; *cp; cp++)
-		      if (!(*cp == ' ' || (*cp >='0' && *cp <= '9')))
+		    do {
+		      for (cp = optarg; *cp; cp++)
+			if (!(*cp == ' ' || (*cp >='0' && *cp <= '9')))
+			  break;
+		      if (!*cp)
 			break;
-
-		    if (*cp)
-		      {
-			new->netid = safe_string_alloc(optarg);
-			optarg = comma + 1;
-			if ((comma = strchr(optarg, ',')))
-			  *comma = 0;
-		      }
+		      
+		      new->netid = safe_malloc(sizeof (struct dhcp_netid));
+		      new->netid->net = safe_string_alloc(optarg);
+		      new->netid->next = np;
+		      np = new->netid;
+		      optarg = comma;
+		      if ((comma = strchr(optarg, ',')))
+			*comma++ = 0;
+		    } while (optarg);
 		  }
 		
-		if ((new->opt = atoi(optarg)) == 0)
+		if (!optarg || (new->opt = atoi(optarg)) == 0)
 		  {
 		    option = '?';
-		    problem = "bad dhcp-opt";
-		    if (new->netid)
-		      free(new->netid);
-		    free(new);
-		    break;
+		    problem = "bad dhcp-option";
 		  }
-		
-		daemon->dhcp_opts = new;
-		
-		if (!comma)
-		  break;
-		
-		/* characterise the value */
-		is_addr = is_hex = is_dec = 1;
-		addrs = digs = 1;
-		for (cp = comma+1; *cp; cp++)
-		  if (*cp == ',')
-		    {
-		      addrs++;
-		      is_dec = is_hex = 0;
-		    }
-		  else if (*cp == ':')
-		    {
-		      digs++;
-		      is_dec = is_addr = 0;
-		    }
-		  else if (*cp == '.')
-		    is_dec = is_hex = 0;
-		  else if (!(*cp >='0' && *cp <= '9'))
-		      {
-			is_dec = is_addr = 0;
-			if (!((*cp >='A' && *cp <= 'F') ||
-			      (*cp >='a' && *cp <= 'f')))
-			  is_hex = 0;
-		      }
-		
-		if (is_hex && digs > 1)
+		else if (comma && new->opt == 119)
 		  {
-		    char *p = comma+1, *q, *r;
-		    new->len = digs;
-		    q = new->val = safe_malloc(new->len);
-		    while (*p)
+		    /* dns search, RFC 3397 */
+		    unsigned char *q, *r, *tail;
+		    unsigned char *p = NULL;
+		    int newlen, len = 0;
+		    
+		    optarg = comma;
+		    if ((comma = strchr(optarg, ',')))
+		      *(comma++) = 0;
+
+		    while (optarg && *optarg)
 		      {
-			for (r = p; *r && *r != ':'; r++);
-			if (*r)
+			if (!canonicalise(optarg))
 			  {
-			    if (r != p)
-			      {
-				*r = 0;
-				*(q++) = strtol(p, NULL, 16);
-			      }
-			    p = r+1;
-			  }
-			else
-			  {
-			    if (*p)
-			      *(q++) = strtol(p, NULL, 16);
+			    option = '?';
+			    problem = "bad domain in dhcp-option";
 			    break;
 			  }
+			
+			if (!(r = realloc(p, len + strlen(optarg) + 2)))
+			  die("could not get memory", NULL);
+			p = memmove(r, p, len);
+			
+			q = p + len;
+			
+			/* add string on the end in RFC1035 format */
+			while (*optarg) 
+			  {
+			    char *cp = q++;
+			    int j;
+			    for (j = 0; *optarg && (*optarg != '.'); optarg++, j++)
+			      *q++ = *optarg;
+			    *cp = j;
+			    if (*optarg)
+			      optarg++;
+			  }
+			*q++ = 0;
+			
+			/* Now tail-compress using earlier names. */
+			newlen = q - p;
+			for (tail = p + len; *tail; tail += (*tail) + 1)
+			  for (r = p; r - p < len; r += (*r) + 1)
+			    if (strcmp(r, tail) == 0)
+			      {
+				PUTSHORT((r - p) | 0xc000, tail); 
+				newlen = tail - p;
+				goto end;
+			      }
+		      end:
+			len = newlen;
+		    
+			optarg = comma;
+			if (optarg && (comma = strchr(optarg, ',')))
+			  *(comma++) = 0;
 		      }
+
+		    new->len = len;
+		    new->val = p;
 		  }
-		else if (is_dec)
+		else if (comma)
 		  {
-		    /* Given that we don't know the length,
-		       this appaling hack is the best available */
-		    unsigned int val = atoi(comma+1);
-		    if (val < 256)
+		    /* not option 119 */
+		    /* characterise the value */
+		    is_addr = is_hex = is_dec = 1;
+		    addrs = digs = 1;
+		    for (cp = comma; *cp; cp++)
+		      if (*cp == ',')
+			{
+			  addrs++;
+			  is_dec = is_hex = 0;
+			}
+		      else if (*cp == ':')
+			{
+			  digs++;
+			  is_dec = is_addr = 0;
+			}
+		      else if (*cp == '.')
+			is_dec = is_hex = 0;
+		      else if (!((*cp >='0' && *cp <= '9') || *cp == '-'))
+			{
+			  is_dec = is_addr = 0;
+			  if (!((*cp >='A' && *cp <= 'F') ||
+				(*cp >='a' && *cp <= 'f')))
+			    is_hex = 0;
+			}
+		
+		    if (is_hex && digs > 1)
 		      {
-			new->len = 1;
-			new->val = safe_malloc(1);
-			*(new->val) = val;
+			char *p = comma, *q, *r;
+			new->len = digs;
+			q = new->val = safe_malloc(new->len);
+			while (*p)
+			  {
+			    for (r = p; *r && *r != ':'; r++);
+			    if (*r)
+			      {
+				if (r != p)
+				  {
+				    *r = 0;
+				    *(q++) = strtol(p, NULL, 16);
+				  }
+				p = r+1;
+			      }
+			    else
+			      {
+				if (*p)
+				  *(q++) = strtol(p, NULL, 16);
+				break;
+			      }
+			  }
 		      }
-		    else if (val < 65536)
+		    else if (is_dec)
 		      {
-			new->len = 2;
-			new->val = safe_malloc(2);
-			*(new->val) = val>>8;
-			*(new->val+1) = val;
+			int i, val = atoi(comma);
+			/* assume numeric arg is 1 byte except for
+			   options where it is known otherwise. */
+			switch (new->opt)
+			  {
+			  default:
+			    new->len = 1;
+			    break;
+			  case 13: case 22: case 25: case 26: 
+			    new->len = 2;
+			    break;
+			  case 2: case 24: case 35: case 38: 
+			    new->len = 4;
+			    break;
+			  }
+			new->val = safe_malloc(new->len);
+			for (i=0; i<new->len; i++)
+			  new->val[i] = val>>((new->len - i - 1)*8);
+		      }
+		    else if (is_addr)	
+		      {
+			struct in_addr in;
+			unsigned char *op;
+			new->len = INADDRSZ * addrs;
+			new->val = op = safe_malloc(new->len);
+			new->is_addr = 1;
+			while (addrs--) 
+			  {
+			    cp = comma;
+			    if ((comma = strchr(cp, ',')))
+			      *comma++ = 0;
+			    in.s_addr = inet_addr(cp);
+			    memcpy(op, &in, INADDRSZ);
+			    op += INADDRSZ;
+			  }
 		      }
 		    else
 		      {
-			new->len = 4;
-			new->val = safe_malloc(4);
-			*(new->val) = val>>24;
-			*(new->val+1) = val>>16;
-			*(new->val+2) = val>>8;
-			*(new->val+3) = val;
+			/* text arg */
+			new->len = strlen(comma);
+			new->val = safe_malloc(new->len);
+			memcpy(new->val, comma, new->len);
 		      }
 		  }
-		else if (is_addr)	
+
+		if (new->len > 256)
 		  {
-		    struct in_addr in;
-		    unsigned char *op;
-		    new->len = INADDRSZ * addrs;
-		    new->val = op = safe_malloc(new->len);
-		    new->is_addr = 1;
-		    while (addrs--) 
-		      {
-			cp = comma;
-			if ((comma = strchr(cp+1, ',')))
-			  *comma = 0;
-			in.s_addr = inet_addr(cp+1);
-			memcpy(op, &in, INADDRSZ);
-			op += INADDRSZ;
-		      }
+		    option = '?';
+		    problem = "dhcp-option too long";
+		  }
+
+		if (option == '?')
+		  {
+		    if (new->netid)
+		      free(new->netid);
+		    if (new->val)
+		      free(new->val);
+		    free(new);
 		  }
 		else
-		  {
-		    /* text arg */
-		    new->len = strlen(comma+1);
-		    new->val = safe_malloc(new->len);
-		    memcpy(new->val, comma+1, new->len);
-		  }
+		  daemon->dhcp_opts = new;
+
 		break;
 	      }
 
 	    case 'M':
 	      {
-		char *comma;
-		
-		if ((comma = strchr(optarg, ',')))
-		  *comma = 0;
-		daemon->dhcp_file = safe_string_alloc(optarg);
-		if (comma)
+		struct dhcp_netid *id = NULL;
+		while (optarg && strstr(optarg, "net:") == optarg)
 		  {
-		    optarg = comma+1;
+		    struct dhcp_netid *newid = safe_malloc(sizeof(struct dhcp_netid));
+		    newid->next = id;
+		    id = newid;
 		    if ((comma = strchr(optarg, ',')))
-		      *comma = 0;
-		    daemon->dhcp_sname = safe_string_alloc(optarg);
-		    if (comma && (daemon->dhcp_next_server.s_addr = inet_addr(comma+1)) == (in_addr_t)-1)
-		      option = '?';
+		      *comma++ = 0;
+		    newid->net = safe_string_alloc(optarg+4);
+		    optarg = comma;
+		  };
+		
+		if (!optarg)
+		  option = '?';
+		else 
+		  {
+		    char *dhcp_file, *dhcp_sname = NULL;
+		    struct in_addr dhcp_next_server;
+		    if ((comma = strchr(optarg, ',')))
+		      *comma++ = 0;
+		    dhcp_file = safe_string_alloc(optarg);
+		    dhcp_next_server.s_addr = 0;
+		    if (comma)
+		      {
+			optarg = comma;
+			if ((comma = strchr(optarg, ',')))
+			  *comma++ = 0;
+			dhcp_sname = safe_string_alloc(optarg);
+			if (comma && (dhcp_next_server.s_addr = inet_addr(comma)) == (in_addr_t)-1)
+			  option = '?';
+		      }
+		    if (option != '?')
+		      {
+			struct dhcp_boot *new = safe_malloc(sizeof(struct dhcp_boot));
+			new->file = dhcp_file;
+			new->sname = dhcp_sname;
+			new->next_server = dhcp_next_server;
+			new->netid = id;
+			new->next = daemon->boot_config;
+			daemon->boot_config = new;
+		      }
+		  }
+
+		if (option == '?')
+		  {
+		    struct dhcp_netid *tmp;
+		    for (; id; id = tmp)
+		      {
+			tmp = id->next;
+			free(id);
+		      }
 		  }
 		break;
 	      }
@@ -1168,8 +1318,6 @@ struct daemon *read_opts (int argc, char **argv)
 	    case 'U':
 	    case 'j':
 	      {
-		char *comma;
-		
 		if (!(comma = strchr(optarg, ',')))
 		  option = '?';
 		else
@@ -1186,7 +1334,27 @@ struct daemon *read_opts (int argc, char **argv)
 		  }
 		break;
 	      }
-		    
+	      
+	    case 'J':
+	      {
+		struct dhcp_netid_list *new = safe_malloc(sizeof(struct dhcp_netid_list));
+		struct dhcp_netid *list = NULL;
+		new->next = daemon->dhcp_ignore;
+		daemon->dhcp_ignore = new;
+		do {
+		  struct dhcp_netid *member = safe_malloc(sizeof(struct dhcp_netid));
+		  if ((comma = strchr(optarg, ',')))
+		    *comma++ = 0;
+		  member->next = list;
+		  list = member;
+		  member->net = safe_string_alloc(optarg);
+		  optarg = comma;
+		} while (optarg);
+		
+		new->list = list;
+		break;
+	      }
+
 	    case 'V':
 	      {
 		char *a[3] = { NULL, NULL, NULL };
@@ -1224,6 +1392,84 @@ struct daemon *read_opts (int argc, char **argv)
 		
 		break;
 	      }
+
+	    case 'W':
+	      {
+		int port = 1, priority = 0, weight = 0;
+		char *name, *target = NULL;
+		struct srv_record *new;
+		
+		if ((comma = strchr(optarg, ',')))
+		  *(comma++) = 0;
+
+		if (!canonicalise(optarg))
+		  {
+		    option = '?';
+		    problem = "bad SRV record";
+		    break;
+		  }
+		name = safe_string_alloc(optarg);
+		
+		if (comma)
+		  {
+		    optarg = comma;
+		    if ((comma = strchr(optarg, ',')))
+		      *(comma++) = 0;
+		    if (!canonicalise(optarg))
+		      {
+			option = '?';
+			problem = "bad SRV target";
+			break;
+		      }
+		    target = safe_string_alloc(optarg);
+		    if (comma)
+		      {
+			optarg = comma;
+			if ((comma = strchr(optarg, ',')))
+			  *(comma++) = 0;
+			if (!atoi_check(optarg, &port))
+			  {
+			    option = '?';
+			    problem = "invalid port number";
+			    break;
+			  }
+			if (comma)
+			  {
+			    optarg = comma;
+			    if ((comma = strchr(optarg, ',')))
+			      *(comma++) = 0;
+			    if (!atoi_check(optarg, &priority))
+			      {
+				option = '?';
+				problem = "invalid priority";
+				break;
+			      }
+			    if (comma)
+			      {
+				optarg = comma;
+				if ((comma = strchr(optarg, ',')))
+				  *(comma++) = 0;
+				if (!atoi_check(optarg, &weight))
+				  {
+				    option = '?';
+				    problem = "invalid weight";
+				    break;
+				  }
+			      }
+			  }
+		      }
+		  }
+		
+		new = safe_malloc(sizeof(struct srv_record));
+		new->next = daemon->srvnames;
+		daemon->srvnames = new;
+		new->srvname = name;
+		new->srvtarget = target;
+		new->srvport = port;
+		new->priority = priority;
+		new->weight = weight;
+		break;
+	      }
 	    }
 	}
       
@@ -1236,7 +1482,11 @@ struct daemon *read_opts (int argc, char **argv)
 	      complain(buff, NULL);
 	    }
 	  else
+#ifdef HAVE_GETOPT_LONG
 	    die("bad command line options: %s.", problem ? problem : "try --help");
+#else
+	    die("bad command line options: %s.", problem ? problem : "try -w");
+#endif
 	}
     }
       
@@ -1268,24 +1518,39 @@ struct daemon *read_opts (int argc, char **argv)
 #endif /* IPv6 */
     }
 		      
-  /* only one of these need be specified: the other defaults to the
-     host-name */
+  /* only one of these need be specified: the other defaults to the host-name */
   if ((daemon->options & OPT_LOCALMX) || daemon->mxnames || daemon->mxtarget)
     {
       if (gethostname(buff, MAXDNAME) == -1)
 	die("cannot get host-name: %s", NULL);
-	      
+      
       if (!daemon->mxnames)
 	{
 	  daemon->mxnames = safe_malloc(sizeof(struct mx_record));
 	  daemon->mxnames->next = NULL;
 	  daemon->mxnames->mxtarget = NULL;
 	  daemon->mxnames->mxname = safe_string_alloc(buff);
-}
+	}
       
       if (!daemon->mxtarget)
 	daemon->mxtarget = safe_string_alloc(buff);
     }
+
+  if (daemon->domain_suffix)
+    {
+       /* add domain for any srv record without one. */
+      struct srv_record *srv;
+      
+      for (srv = daemon->srvnames; srv; srv = srv->next)
+	if (strchr(srv->srvname, '.') && strchr(srv->srvname, '.') == strrchr(srv->srvname, '.'))
+	  {
+	    strcpy(buff, srv->srvname);
+	    strcat(buff, ".");
+	    strcat(buff, daemon->domain_suffix);
+	    free(srv->srvname);
+	    srv->srvname = safe_string_alloc(buff);
+	  }
+    }
   
   if (daemon->options & OPT_NO_RESOLV)
     daemon->resolv_files = 0;
